﻿{
//INSTANT 3D SCENE
// -- Created by Nat Russell at DesignProTools.com - 8-2-19 --
//Drop this script in the 'Script UI' folder in the 'Scripts' folder in your After Effects application folder
//This is a free script, and if you have any issues, or for a more advanced script go buy PT_Multiplane (https://aescripts.com/pt_multiplane/)
//Multiplane's concept was the inspiration for the below script, make sure to get his full one after using this. Make sure to go to Design Pro Tools for more free tools and tutorials.

function instant3D(thisObj) {
    function instant3D_buildUI(thisObj) {
        var myPanel = (thisObj instanceof Panel) ? thisObj : new Window("palette", "", [100, 100, 300, 280]);//

        //Create the panel, elements and buttons
        var firstDistanceLabel =  myPanel.add("statictext", [10,10,120,30], "First layer distance");
        var firstDistance =  myPanel.add("edittext", [125,10,180,30], "0");
        var lastDistanceLabel =  myPanel.add("statictext", [10,35,120,55], "Last layer distance");
        var lastDistance =  myPanel.add("edittext", [125,35,180,55], "4000");
        var ddList = new Array("Distribute Layers Evenly", "Random Layer Position");
        var distributeDropdown = myPanel.add("dropdownlist", [10,60,180,80], ddList);
            distributeDropdown.selection = 0;
        var makeSceneButton = myPanel.add("button", [10,85,180,125], "Make 3D Scene");
        var textDescription = myPanel.add("statictext", [10,130,180,180], "Move layers in Z after above.\nHit below to remove expressions\nand keep z positions.",{multiline:true});
        var finalisePositionButton = myPanel.add("button", [10,185,140,205], "Finalise Position");
        var websiteButton = myPanel.add("button", [150,185,180,205], "?");

    //when they click 'Make 3d Scene Button'
    makeSceneButton.onClick = function() {
        //pull the options selected from the panel and call the main function
        instant3DScene(parseInt(firstDistance.text,10),parseInt(lastDistance.text,10),distributeDropdown.selection);
    }

    //when they click 'finalise position' button
    finalisePositionButton.onClick = function() {
        //pull the options selected from the panel and call the main function
        finalisePosition();
    }

    //when they clicked ? button go to the website
    websiteButton.onClick = function() {
        if (app.preferences.getPrefAsLong("Main Pref Section", "Pref_SCRIPTING_FILE_NETWORK_SECURITY") != 1)
        {
            alert("Please tick the \"Allow Scripts to Write Files and Access Network\" checkbox if Preferences > General");
            // Then open Preferences > General to let people tick the checkbox
            app.executeCommand(2359);
        }
        var os = system.osName;
        if (!os.length) {os = $.os;}
        app_os =  ( os.indexOf("Win") != -1 )  ?  "Win" : "Mac"
        var url = "http://designprotools.com";
        if ( app_os == "Win" )
        {
            system.callSystem("explorer " + url);
        }
        else
        {
            system.callSystem("open " + url);
        }
    }

        return myPanel;
    } //END OF PANEL CREATION

/// THIS WILL CHECK IF PANEL IS DOCKABLE OR FLAOTING WINDOW
    var instant3DPal = instant3D_buildUI(thisObj);
    if ((instant3DPal != null) && (instant3DPal instanceof Window)) {
        instant3DPal.center();
        instant3DPal.show();
        }
    }
//create panel
instant3D(this);


//Function that does all the actual work. This will get all the layers that you want and put lay them out in 3D space brother
function instant3DScene(startDistance, endDistance, random) {
    // Create script undo group
    app.beginUndoGroup("Instant 3D scene");
    //////now get the comp
    var comp = getActiveComp();
    var myLayers = comp.layers;
    var mySelectedLayers = comp.selectedLayers;
    var chosenLayers = [];
    var errorMSG = "";

    //if there is 1 or no layers selected (so it looks like they have not selected the layers to perform this to) then do it do all layers
    if (mySelectedLayers.length <= 1) {
        for (var i = 1; i <= myLayers.length; i++) {
            //the array with all layers go from 1 upwards, so put into an array so it starts at 0 (like the selected array)
            chosenLayers.push(myLayers[i]);
        } 
    }else {
            chosenLayers = mySelectedLayers;
    }
    

    //figure out distance between the layers
    var distanceBetween = (endDistance - startDistance)/chosenLayers.length;
    //randomise layers if they asked for it
    if (random == 1) {
        randomise(chosenLayers);
    }
        //loop through the selected layers and do the things
        for (var i = 0; i < chosenLayers.length; i++) {
            if(chosenLayers[i].property("Scale").canSetExpression) {
                chosenLayers[i].threeDLayer = true;

                //clear anchor point, position and scale keyframes if they are there
                var currentProp = chosenLayers[i].property("Position");
                if(currentProp.numKeys>0){
                    for (i = currentProp.numKeys; i != 0; i--) { // Cycle through the Keyframes
                        currentProp.removeKey(i); // Remove the current Keyframe
                    }
                    errorMSG = "Position, Scale and Anchor Point keyframes are removed to get expressions to work. Hit undo after this alert to get them back and cancel script.";
                }
                currentProp = chosenLayers[i].property("Scale");
                if(currentProp.numKeys>0){
                    for (i = currentProp.numKeys; i != 0; i--) { // Cycle through the Keyframes
                        currentProp.removeKey(i); // Remove the current Keyframe
                    }
                    errorMSG = "Position, Scale and Anchor Point keyframes are removed to get expressions to work. Hit undo after this alert to get them back and cancel script.";
                }
                currentProp = chosenLayers[i].property("Anchor Point");
                if(currentProp.numKeys>0){
                    for (i = currentProp.numKeys; i != 0; i--) { // Cycle through the Keyframes
                        currentProp.removeKey(i); // Remove the current Keyframe
                    }
                    errorMSG = "Position, Scale and Anchor Point keyframes are removed to get expressions to work. Hit undo after this alert to get them back and cancel script.";
                }


                //set anchor point to center, so scaling works. When you hit the finalise button, it will put it to the center of the layer
                //get current anchor point value
                var curAn = chosenLayers[i].property("Anchor Point").value;
                //get current position of layer
                var curPos = chosenLayers[i].property("Position").value;
                //set anchor point to 0, so it's in the center of screen
                chosenLayers[i].property("Anchor Point").setValue([((comp.width/2)-curPos[0])+curAn[0],((comp.height/2)-curPos[1])+curAn[1],0]);
                chosenLayers[i].property("Position").setValue([comp.width/2,comp.height/2,startDistance]);
                
                chosenLayers[i].property("Scale").expression = 'c = thisComp.layer("Camera"); point1 = toWorld(anchorPoint); point2 = c.toWorld([0,0,0]); d = length(point1, point2); s = (d/c.zoom)*100; [s, s, 100]';
                startDistance +=distanceBetween;
                //add a marker that lets you know it's a live 3D scene
                chosenLayers[i].property("Marker").setValueAtTime(comp.time, new MarkerValue("Instant 3D scene live. Move z position."));
            }
        }

    //create camera and null
    var myCam = myLayers.addCamera("Camera", [0,0]);
    var negZoom = -myCam.property("ADBE Camera Options Group").property("ADBE Camera Zoom").value;
    myCam.property("ADBE Transform Group").property("ADBE Position").setValue([comp.width/2, comp.height/2,negZoom]);
    myCam.property("ADBE Transform Group").property("ADBE Anchor Point").setValue([comp.width/2, comp.height/2,0]);
    var camControl = comp.layers.addNull();
    camControl.name = "Camera Control";
    camControl.threeDLayer = true;
    myCam.parent = comp.layers.byName("Camera Control");

    //alert message if there was an issue
    if (errorMSG != "") alert(errorMSG);

    app.endUndoGroup();
}

//Function for randomising the layers if they want it, which isn't often
function randomise(array) {
      //array starts normally, at 0 (this is the selected layers)
      var currentIndex = array.length, temporaryValue, randomIndex;
      // While there remain elements to shuffle...
      while (0 !== currentIndex) {
        // Pick a remaining element...
        randomIndex = Math.floor(Math.random() * currentIndex);
        currentIndex -= 1;
        // And swap it with the current element.
        temporaryValue = array[currentIndex];
        array[currentIndex] = array[randomIndex];
        array[randomIndex] = temporaryValue;
      }

  return array;
}

function finalisePosition() {
    app.beginUndoGroup("Finalise Position");
    //////now get the comp
    var comp = getActiveComp();
    var myLayers = comp.layers;
    var errorMSG = "";
    for (var i = 1; i <= myLayers.length; i++) {
        //remove keyframes of scale before cleaning
        currentProp = myLayers[i].property("Scale");
        if(currentProp.numKeys>0){
            for (j = currentProp.numKeys; j != 0; j--) { // Cycle through the Keyframes
                currentProp.removeKey(j); // Remove the current Keyframe
            }
            errorMSG = "Scale keyframes have been removed to get finalised scale. Hit undo after this alert to get them back and cancel finalisation.";
        }
        //remove expression on scale
        if(myLayers[i].property("Scale").canSetExpression) {
            //store the scale based on expression to add after removing expression
            var tempScale = myLayers[i].property("Scale").value;
            myLayers[i].property("Scale").expression = '';
            //add on scale before removing expression
            myLayers[i].property("Scale").setValue(tempScale);
        }
        
        //remove the first marker on the layer
        var myMarkers = myLayers[i].marker;
        if(myMarkers.numKeys > 0) myMarkers.removeKey(1);

    } 

    //alert message if there was an issue
    if (errorMSG != "") alert(errorMSG);
    app.endUndoGroup();
}


//////These 2 functions are just for getting the selected comp
function activateCompViewer(){  
    // setActive is supposed (guide) to return a Boolean, but in practice it returns nothing, therefore this doesnt work:  
    // return app.activeViewer && app.activeViewer.type===ViewerType.VIEWER_COMPOSITION && app.activeViewer.setActive();  
     
    var A = (app.activeViewer && app.activeViewer.type===ViewerType.VIEWER_COMPOSITION);  
    if (A) app.activeViewer.setActive();  
    return A;  
    };  
function getActiveComp(){  
     
    var comp;                                // the returned quantity  
    var X = app.project.activeItem;    // the initial activeItem  
    var selComp = app.project.selection.length ===1 && app.project.selection[0].typeName === "Composition" ? app.project.selection[0] : null; // the unique selected comp, or null  
    var temp;  
    if (X instanceof CompItem){  
        if (selComp === null){  
            comp = X;  
            }  
        else if (selComp !== X){  
            comp = null; // ambiguity : the timeline panel is active, X is the front comp, but another comp is selected  
            }  
        else{  
            // X and selComp coincide  
            X.selected = false;  
            temp = app.project.activeItem;  
            X.selected = true;  
            if (temp === null){  
                // the project panel is active and the active item was initially a selected comp  
                // if that comp is already opened in a viewer and is the front comp, return the comp, else : ambiguity  
                comp = (activateCompViewer() && app.project.activeItem === X) ? X : null;  
                }  
            else{  
                // deselecting the comp didn't change the activeItem : the timeline panel is active, and the active item was the front comp, that happened to be also selected.  
                comp = X;  
                };  
            };  
        }  
    else{  
        comp = activateCompViewer() ? app.project.activeItem : null;  
        };  
    return comp;  
};  

}